
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/wait.h>

#include <event.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "xpw.h"
#include "xpwd.h"

extern int	local_chpass(struct xpw *, struct passwd *);
extern int	local_passwd(struct xpw *, struct passwd *);

int
xpwd_open(struct xpwd_ctx *ctx)
{	
	struct sockaddr_un sun;
	
	memset(&sun, 0, sizeof(sun));
	sun.sun_len = sizeof(sun);
	sun.sun_family = AF_UNIX;
	strlcpy(sun.sun_path, ctx->sockname, sizeof(sun.sun_path));

	unlink(ctx->sockname);
	
	if ((ctx->fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		return (-1);
	
	if (bind(ctx->fd, (struct sockaddr *)&sun, sizeof(sun)) < 0 ||
	    chmod(ctx->sockname, ACCESSPERMS) < 0 || listen(ctx->fd, 10) < 0) {
		xpwd_close(ctx);
		return (-1);
	}
	event_set(&ctx->recv_ev, ctx->fd, EV_READ|EV_PERSIST,
	    xpwd_recv, ctx);
	event_add(&ctx->recv_ev, NULL);
	
	signal_set(&ctx->int_ev, SIGINT, xpwd_signal, ctx);
	signal_add(&ctx->int_ev, NULL);
	signal_set(&ctx->term_ev, SIGTERM, xpwd_signal, ctx);
	signal_add(&ctx->term_ev, NULL);
	signal_set(&ctx->chld_ev, SIGCHLD, xpwd_signal, ctx);
	signal_add(&ctx->chld_ev, NULL);

	return (0);
}

void
xpwd_close(struct xpwd_ctx *ctx)
{
	if (ctx->fd != -1) {
		unlink(ctx->sockname);
		close(ctx->fd);
		ctx->fd = -1;
	}
	event_del(&ctx->recv_ev);
	event_del(&ctx->chld_ev);
	event_del(&ctx->int_ev);
	event_del(&ctx->term_ev);
}

static void
xpwd_child(int fd)
{
	struct xpw req, rep;
	struct passwd *pw;
	char buf[4096];
	uid_t uid;
	gid_t gid;
	int len;

	/* Identify remote client. */
	if (getpeereid(fd, &uid, &gid) < 0) {
		syslog(LOG_ERR, "getpeereid failed: %m");
		return;
	}
	if ((pw = getpwuid(uid)) == NULL) {
		syslog(LOG_ERR, "connection from unknown uid %u", uid);
		return;
	}
	/* Read client request. */
	if ((len = read(fd, buf, sizeof(buf))) < 0 ||
	    xpw_xdr_decode(buf, len, &req) < 0) {
		syslog(LOG_ERR, "error reading request from %s: %m",
		    pw->pw_name);
		return;
	}
	/* Execute client request. */
	switch (req.xpw_type) {
	case XPW_REQ_PASSWD:
		if (local_passwd(&req, pw) == 0) {
			syslog(LOG_INFO, "password change succeeded for %s",
			    pw->pw_name);
			xpw_xdr_free(&req);
			rep.xpw_type = XPW_REP_OK;
			rep.xpw_reply = "Password changed.";
		} else {
			syslog(LOG_ERR, "password change failed for %s",
			    pw->pw_name);
			xpw_xdr_free(&req);
			rep.xpw_type = XPW_REP_ERR;
			rep.xpw_reply = "Permission denied.";
		}
		break;
	case XPW_REQ_CHPASS:
		if (local_chpass(&req, pw) == 0) {
			syslog(LOG_INFO, "user information change succeeded "
			    "for %s", pw->pw_name);
			xpw_xdr_free(&req);
			rep.xpw_type = XPW_REP_OK;
			rep.xpw_reply = "User information changed.";
		} else {
			syslog(LOG_ERR, "user information change failed "
			    "for %s", pw->pw_name);
			xpw_xdr_free(&req);
			rep.xpw_type = XPW_REP_ERR;
			rep.xpw_reply = "Permission denied.";
		}
		break;
	default:
		syslog(LOG_ERR, "unknown request %d by %s", req.xpw_type,
		    pw->pw_name);
		xpw_xdr_free(&req);
		rep.xpw_type = XPW_REP_ERR;
		rep.xpw_reply = "Bad request.";
		break;
	}
	/* Return results. */
	len = xpw_xdr_encode(&rep, buf, sizeof(buf));
	
	if (write(fd, buf, len) != len)
		syslog(LOG_ERR, "couldn't reply to uid %u: %m", uid);
}

void
xpwd_recv(int fd, short event, void *arg)
{
	struct xpwd_ctx *ctx = (struct xpwd_ctx *)arg;
	struct sockaddr_un sun;
	socklen_t slen = sizeof(sun);
	int cfd;

	if (ctx->children < 100 &&
	    (cfd = accept(ctx->fd, (struct sockaddr *)&sun, &slen)) != -1) {
		if (fork() == 0) {
			close(ctx->fd);
			ctx->fd = -1;
			xpwd_close(ctx);

			/* Handle child request. */
			xpwd_child(cfd);
			
			exit(0);
		}
		close(cfd);
		ctx->children++;
	}
}

void
xpwd_signal(int sig, short event, void *arg)
{
	struct xpwd_ctx *ctx = (struct xpwd_ctx *)arg;
	pid_t pid, status;
	
	switch (sig) {
	case SIGCHLD:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
			if (!WIFEXITED(status))
				syslog(LOG_ERR, "child (pid %u) exited with "
				    "status %d", pid, WEXITSTATUS(status));
			ctx->children--;
		}
		break;
	default:
		syslog(LOG_INFO, "terminating on signal %d", sig);
		xpwd_close(ctx);
		exit(0);
	}
}
